"""Constants for unit tests."""

import json
import pathlib

from yarl import URL

RESOURCES_ROOT = pathlib.Path(__file__).parent / "resources"

RAW_GITHUB_RELEASES = (RESOURCES_ROOT / "github_releases.json").read_text()
RAW_MANIFEST_JSON = (RESOURCES_ROOT / "manifest.json").read_text()

GITHUB_RELEASES = json.loads(RAW_GITHUB_RELEASES)
MANIFEST_JSON = json.loads(RAW_MANIFEST_JSON)

# Use the latest stable release (v2025.09.30 at index 1, after prerelease)
GITHUB_API_RESPONSE = GITHUB_RELEASES[1]

assert GITHUB_API_RESPONSE["assets"][0]["name"] == "manifest.json"

MANIFEST_URL = URL(GITHUB_API_RESPONSE["assets"][0]["browser_download_url"])
MANIFEST_HTML_URL = URL(GITHUB_API_RESPONSE["html_url"])



"""Firmware update client."""

from .client import FirmwareUpdateClient, ManifestMissing
from .models import FirmwareManifest, FirmwareMetadata

__all__ = [
    "FirmwareUpdateClient",
    "FirmwareManifest",
    "FirmwareMetadata",
    "ManifestMissing",
]



import copy
import dataclasses
import json

from aiohttp import ClientSession
from aioresponses import aioresponses
import pytest
from yarl import URL

from ha_silabs_firmware_client.client import FirmwareUpdateClient, ManifestMissing

from .const import GITHUB_API_RESPONSE, GITHUB_RELEASES, RESOURCES_ROOT

API_URL = "https://api.github.com/repos/NabuCasa/silabs-firmware-builder/releases"


async def test_firmware_update_client() -> None:
    """Test the firmware update client loads manifests."""
    async with ClientSession() as session:
        with aioresponses() as http:
            # Mock all assets
            http.get(API_URL, body=json.dumps(GITHUB_RELEASES))

            for asset in GITHUB_API_RESPONSE["assets"]:
                assert (RESOURCES_ROOT / asset["name"]).is_relative_to(RESOURCES_ROOT)
                http.get(
                    asset["browser_download_url"],
                    body=(RESOURCES_ROOT / asset["name"]).read_bytes(),
                )

            client = FirmwareUpdateClient(API_URL, session)
            manifest = await client.async_update_data()

            assert manifest.url == URL(
                "https://github.com/NabuCasa/silabs-firmware-builder/releases/download/v2025.09.30/manifest.json"
            )
            assert manifest.html_url == URL(
                "https://github.com/NabuCasa/silabs-firmware-builder/releases/tag/v2025.09.30"
            )
            assert len(manifest.firmwares) == 10

            # All firmwares validate
            for fw in manifest.firmwares:
                data = await client.async_fetch_firmware(fw)
                assert isinstance(data, bytes)

            # Load things again
            http.get(API_URL, body=json.dumps(GITHUB_RELEASES))

            new_manifest = await client.async_update_data()
            assert manifest is new_manifest

            # Because the cached URL did not change, we did not need to download again


async def test_firmware_update_client_manifest_missing() -> None:
    """Test the firmware update client handles missing manifests."""
    releases = copy.deepcopy(GITHUB_RELEASES)
    releases[1]["assets"] = [
        a for a in releases[1]["assets"] if a["name"] != "manifest.json"
    ]

    async with ClientSession() as session:
        with aioresponses() as http:
            http.get(API_URL, body=json.dumps(releases))
            client = FirmwareUpdateClient(API_URL, session)

            with pytest.raises(ManifestMissing):
                await client.async_update_data()


async def test_fetch_firmware() -> None:
    """Test fetching firmware."""
    async with ClientSession() as session:
        with aioresponses() as http:
            http.get(API_URL, body=json.dumps(GITHUB_RELEASES))

            for asset in GITHUB_API_RESPONSE["assets"]:
                assert (RESOURCES_ROOT / asset["name"]).is_relative_to(RESOURCES_ROOT)
                http.get(
                    asset["browser_download_url"],
                    body=(RESOURCES_ROOT / asset["name"]).read_bytes(),
                    repeat=True,
                )

            client = FirmwareUpdateClient(API_URL, session)
            manifest = await client.async_update_data()

            for meta in manifest.firmwares:
                data = await client.async_fetch_firmware(meta)
                assert isinstance(data, bytes)

            # Invalid firmwares are caught during fetching
            meta = dataclasses.replace(
                manifest.firmwares[0],
                checksum="sha3-256:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            )

            with pytest.raises(ValueError, match="Invalid firmware checksum"):
                await client.async_fetch_firmware(meta)


async def test_update_prerelease_flag() -> None:
    """Test that update_prerelease() correctly toggles between stable and prerelease."""
    async with ClientSession() as session:
        with aioresponses() as http:
            # Mock all assets for both stable and prerelease versions
            http.get(API_URL, body=json.dumps(GITHUB_RELEASES), repeat=True)

            for asset in GITHUB_API_RESPONSE["assets"]:
                http.get(
                    asset["browser_download_url"],
                    body=(RESOURCES_ROOT / asset["name"]).read_bytes(),
                    repeat=True,
                )

            prerelease = GITHUB_RELEASES[0]
            for asset in prerelease["assets"]:
                http.get(
                    asset["browser_download_url"],
                    body=(RESOURCES_ROOT / asset["name"]).read_bytes(),
                    repeat=True,
                )

            # Start with prerelease=False (default)
            client = FirmwareUpdateClient(API_URL, session)
            manifest = await client.async_update_data()

            # Should get stable release
            assert manifest.html_url == URL(
                "https://github.com/NabuCasa/silabs-firmware-builder/releases/tag/v2025.09.30"
            )

            # Toggle to prerelease=True
            client.update_prerelease(True)
            manifest = await client.async_update_data()

            # Should now get prerelease
            assert manifest.html_url == URL(
                "https://github.com/NabuCasa/silabs-firmware-builder/releases/tag/v2025.10.14"
            )

            # Toggle back to prerelease=False
            client.update_prerelease(False)
            manifest = await client.async_update_data()

            # Should get stable release again
            assert manifest.html_url == URL(
                "https://github.com/NabuCasa/silabs-firmware-builder/releases/tag/v2025.09.30"
            )



"""Firmware update models."""

from dataclasses import dataclass
from datetime import datetime
import hashlib
from typing import Any, Self

from yarl import URL


@dataclass(frozen=True)
class FirmwareMetadata:
    """Metadata for a remotely hosted firmware file."""

    filename: str
    checksum: str
    size: int
    release_notes: str | None
    metadata: dict[str, str | int | None]
    url: URL
    release_summary: str | None = None

    @classmethod
    def from_json(cls, data: dict[str, Any], *, url: URL) -> Self:
        """Construct from JSON data."""
        return cls(
            filename=data["filename"],
            checksum=data["checksum"],
            size=data["size"],
            release_notes=data["release_notes"],
            metadata=data["metadata"],
            # The manifest does not contain the full URL so we pass it externally
            url=url,
            release_summary=data.get("release_summary"),
        )

    def as_dict(self) -> dict[str, Any]:
        """Return metadata as a dict."""
        return {
            "filename": self.filename,
            "checksum": self.checksum,
            "size": self.size,
            "release_notes": self.release_notes,
            "metadata": self.metadata,
            "url": str(self.url),
            "release_summary": self.release_summary,
        }

    def validate_firmware(self, data: bytes) -> None:
        """Parse firmware bytes into a firmware image."""
        if len(data) != self.size:
            raise ValueError("Invalid firmware size")

        algorithm, _, digest = self.checksum.partition(":")
        hasher = hashlib.new(algorithm)
        hasher.update(data)

        if hasher.hexdigest() != digest:
            raise ValueError("Invalid firmware checksum")


@dataclass(frozen=True)
class FirmwareManifest:
    """Manifest for a group of firmwares encompassing a firmware builder release."""

    url: URL
    html_url: URL
    created_at: datetime
    firmwares: tuple[FirmwareMetadata, ...]

    @classmethod
    def from_json(
        cls,
        data: dict[str, Any],
        *,
        url: URL,
        html_url: URL,
    ) -> Self:
        """Construct from JSON data."""
        return cls(
            url=url,
            html_url=html_url,
            created_at=datetime.fromisoformat(data["metadata"]["created_at"]),
            firmwares=tuple(
                [
                    FirmwareMetadata.from_json(f, url=url.parent / f["filename"])
                    for f in data["firmwares"]
                ]
            ),
        )

    def as_dict(self) -> dict[str, Any]:
        """Return manifest as a dict."""
        return {
            "url": str(self.url),
            "html_url": str(self.html_url),
            "metadata": {
                "created_at": self.created_at.isoformat(),
            },
            "firmwares": [f.as_dict() for f in self.firmwares],
        }



"""Firmware update client."""

from __future__ import annotations

import asyncio
import logging

from aiohttp import ClientSession
from packaging.version import Version
from yarl import URL

from .models import FirmwareManifest, FirmwareMetadata

_LOGGER = logging.getLogger(__name__)


class ManifestMissing(Exception):
    """Manifest is missing from the GitHub release."""


class FirmwareUpdateClient:
    """Client to manage firmware updates."""

    def __init__(
        self, url: str, session: ClientSession, *, prerelease: bool = False
    ) -> None:
        """Initialize the firmware update client."""
        self.url = url
        self.session = session

        self._prerelease = prerelease
        self._latest_release_url: str | None = None
        self._latest_manifest: FirmwareManifest | None = None

    def update_prerelease(self, prerelease: bool) -> None:
        """Update whether to include prereleases."""
        self._prerelease = prerelease

    async def async_update_data(self) -> FirmwareManifest:
        async with self.session.get(
            self.url,
            headers={"X-GitHub-Api-Version": "2022-11-28"},
            raise_for_status=True,
        ) as rsp:
            releases = await rsp.json()

        if self._prerelease:
            filtered_releases = releases
        else:
            # Ignore prereleases if we aren't opted in
            filtered_releases = [r for r in releases if not r["prerelease"]]

        # Pick the latest release
        sorted_releases = sorted(
            filtered_releases,
            key=lambda r: Version(r["tag_name"]),
            reverse=True,
        )
        obj = sorted_releases[0]

        release_url = obj["html_url"]

        if release_url == self._latest_release_url:
            _LOGGER.debug("GitHub release URL has not changed")
            assert self._latest_manifest is not None
            return self._latest_manifest

        try:
            manifest_asset = next(
                a for a in obj["assets"] if a["name"] == "manifest.json"
            )
        except StopIteration as exc:
            raise ManifestMissing(
                "GitHub release assets haven't been uploaded yet"
            ) from exc

        # Within the metadata, download the `manifest.json` file
        async with self.session.get(
            manifest_asset["browser_download_url"], raise_for_status=True
        ) as rsp:
            manifest_obj = await rsp.json(content_type=None)

        manifest = FirmwareManifest.from_json(
            manifest_obj,
            url=URL(manifest_asset["browser_download_url"]),
            html_url=URL(release_url),
        )

        # Only set the release URL down here to make sure that we don't invalidate
        # future requests if an exception is raised halfway through this method
        self._latest_manifest = manifest
        self._latest_release_url = release_url

        return self._latest_manifest

    async def async_fetch_firmware(self, meta: FirmwareMetadata) -> bytes:
        """Fetch the firmware file."""
        async with self.session.get(meta.url, raise_for_status=True) as rsp:
            data = await rsp.read()

        await asyncio.get_running_loop().run_in_executor(
            None, meta.validate_firmware, data
        )

        return data



import hashlib

import pytest

from ha_silabs_firmware_client.models import FirmwareManifest, FirmwareMetadata

from .const import MANIFEST_HTML_URL, MANIFEST_JSON, MANIFEST_URL


def test_firmware_metadata() -> None:
    """Test FirmwareMetadata JSON serialization/deserialization."""
    firmware_data = MANIFEST_JSON["firmwares"][0]
    meta = FirmwareMetadata.from_json(
        firmware_data, url=MANIFEST_URL.parent / firmware_data["filename"]
    )

    assert meta.as_dict() == {
        **firmware_data,
        "url": meta.as_dict()["url"],
    }

    assert meta.filename == firmware_data["filename"]
    assert meta.checksum == firmware_data["checksum"]
    assert meta.size == firmware_data["size"]
    assert meta.metadata == firmware_data["metadata"]
    assert meta.release_notes == firmware_data["release_notes"]
    assert meta.url == MANIFEST_URL.parent / firmware_data["filename"]


def test_firmware_metadata_validate_firmware() -> None:
    """Test FirmwareMetadata firmware parsing."""
    firmware = b"Test firmware"

    meta = FirmwareMetadata.from_json(
        {
            "filename": "test_firmware.gbl",
            "checksum": f"sha3-256:{hashlib.sha3_256(firmware).hexdigest()}",
            "size": len(firmware),
            "metadata": {
                "baudrate": 115200,
                "ezsp_version": "7.4.4.0",
                "fw_type": "zigbee_ncp",
                "fw_variant": None,
                "metadata_version": 2,
                "sdk_version": "4.4.4",
            },
            "release_notes": None,
            "release_summary": None,
        },
        url=MANIFEST_URL.parent / "test_firmware.gbl",
    )

    meta.validate_firmware(firmware)

    # The firmware size is checked
    with pytest.raises(ValueError, match="Invalid firmware size"):
        meta.validate_firmware(firmware + b"\x00")

    # As is the checksum
    with pytest.raises(ValueError, match="Invalid firmware checksum"):
        meta.validate_firmware(firmware[:-1] + b"\x00")


def test_firmware_manifest() -> None:
    """Test FirmwareManifest."""
    manifest = FirmwareManifest.from_json(
        MANIFEST_JSON, url=MANIFEST_URL, html_url=MANIFEST_HTML_URL
    )

    assert manifest.as_dict() == {
        "metadata": {
            "created_at": MANIFEST_JSON["metadata"]["created_at"],
        },
        "firmwares": [fw.as_dict() for fw in manifest.firmwares],
        "url": str(MANIFEST_URL),
        "html_url": str(MANIFEST_HTML_URL),
    }

    assert manifest.url == MANIFEST_URL
    assert manifest.html_url == MANIFEST_HTML_URL
    assert manifest.created_at.isoformat() == MANIFEST_JSON["metadata"]["created_at"]
    assert len(manifest.firmwares) == 10
    assert manifest.firmwares[0].filename == "skyconnect_zigbee_ncp_7.4.4.3.gbl"
    assert manifest.firmwares[1].filename == "zwave_stick_7.22.2.0.gbl"
